#!/usr/bin/env python
# -*- coding: utf-8 -*0


"""An asynchronous subprocess utility."""


import asyncio
import concurrent.futures
import logging
import typing
from typing import Union


def cast_proc(obj) -> asyncio.subprocess.Process:
    """Cast helper method."""
    if isinstance(obj, asyncio.subprocess.Process):
        return obj
    raise


"""Callback function definition for class SubProcess

Parameters
----------
line : str
    A string of string sub-process outputs.
isstdout : bool
    If the string is for STDOUT, then True.
    If the string is for STDERR, then False.
"""
CallbackPerLine = typing.Callable[[str, bool], None]


"""Callback function definition for class SubProcess

Parameters
----------
line : bytes
    A bytes of string sub-process outputs.
isstdout : bool
    If the string is for STDOUT, then True.
    If the string is for STDERR, then False.
"""
CallbackPerLineBytes = typing.Callable[[bytes, bool], None]


class SubProcess:
    """Execute a sub process and capture stdout and stderr.

    This class wraps :class:`asyncio.subprocess.Process`.

    Set program(path) and the arguments,
    and then call exec() with concurrent.futures functionality.
    You should use with asyncio event loop as::

        # event loop sample code
        import asyncio
        import typing
        import zobepy

        print('event loop start')
        p = zobepy.SubProcess()
        p.program = '/bin/ls'
        p.args = ['-al']
        loop = typing.cast(asyncio.events.AbstractEventLoop, asyncio.get_event_loop())
        loop.run_until_complete(
          p.exec()
        )
        print('event loop end')

    To capture all stdout and stderr in semi-realtime(line based),
    use callback.

    """

    def __init__(self):
        """ctor."""
        self._logger = logging.getLogger(__name__)
        self._log_prefix: str = ''
        self._program: str = ''
        self._args: typing.List[str] = []
        self._pid: int = -1
        self._return_code: Union[int, None] = None
        self._process: Union[asyncio.subprocess.Process, None] = None
        self._callback_per_line: Union[CallbackPerLine, None] = None
        # self._callback_per_line: CallbackPerLine = lambda a='', b=True: None
        self._callback_per_line_bytes: Union[CallbackPerLineBytes, None] = None
        self._callback_per_line_async = None
        self._callback_per_line_bytes_async = None

    def _get_logger(self) -> logging.Logger:
        return self._logger

    def _set_logger(self, logger: logging.Logger):
        self._logger = logger

    logger = property(_get_logger, _set_logger)
    """Get or set(change) logger."""

    def _get_log_prefix(self) -> str:
        return self._log_prefix

    def _set_log_prefix(self, str: str):
        self._log_prefix = str

    log_prefix = property(_get_log_prefix, _set_log_prefix)

    def _get_program(self):
        return self._program

    def _set_program(self, program: str):
        self._program = program

    program = property(_get_program, _set_program)
    """Program path to execute."""

    def _get_args(self):
        return self._args

    def _set_args(self, args):
        self._args = args

    def set_args(self, *args):
        """Set arguments to the program."""
        self._args = args

    args = property(_get_args, _set_args)
    """Arguments to be passed to the program"""

    @property
    def pid(self) -> int:
        """Processid of the executing program process.

        -1 if the process has not been started yet.
        """
        return self._pid

    @pid.setter
    def pid(self, pid: int):
        self._pid = pid

    def _get_return_code(self) -> Union[int, None]:
        return self._return_code

    def _set_return_code(self, return_code: int):
        self._return_code = return_code

    return_code = property(_get_return_code)
    """The return value of the executing program process.

    None if the process is running or before run.
    """

    # def _adapt_stdout(self, line):
    #     self._process_stream(line, True)
    #
    # def _adapt_stderr(self, line):
    #     self._process_stream(line, False)

    @property
    def callback_per_line(self) -> Union[CallbackPerLine, None]:
        """Set callback function to get subprocess output in real time.

        Callback function specification:
            func(line: str, isstdout: bool) -> None

        Set
        callback function::

            def your_function(line: str, isstdout: bool):
                print(line)

            sp = zobepy.SubProcess()
            sp.callback_per_line = your_function


        Lambda function
        is also available::

            sp = zobepy.SubProcess()
            sp.callback_per_line = lambda line, isstdout: print(line)

        Note:
            This function implicitly converts subprocess output bytes
            into string as 'utf-8'.

            Use :func:`callback_per_line_bytes` to get data as-is.

        Warning:
            Do not block for a long time in this function
            to keep condition good about parallel execution.
            Otherwise use :func:`callback_per_line_async` instead.

        """
        return self._callback_per_line

    @callback_per_line.setter
    def callback_per_line(self, callback_per_line: CallbackPerLine):
        self._callback_per_line = callback_per_line

    @property
    def callback_per_line_bytes(self) -> Union[CallbackPerLineBytes, None]:
        """Set callback function to get subprocess output bytes in real time."""
        return self._callback_per_line_bytes

    @callback_per_line_bytes.setter
    def callback_per_line_bytes(self, func: CallbackPerLineBytes):
        self._callback_per_line_bytes = func

    def _get_callback_per_line_async(self):
        return self._callback_per_line_async

    def _set_callback_per_line_async(self, callback_per_line_async):
        if callback_per_line_async is None:
            self._callback_per_line_async = None
        elif asyncio.iscoroutinefunction(callback_per_line_async):
            self._callback_per_line_async = callback_per_line_async
        else:
            raise TypeError('A coroutine function or None is '
                            'required')

    callback_per_line_async = property(
        _get_callback_per_line_async, _set_callback_per_line_async)
    """Set callback coroutine function to get subprocess output bytes.

    """

    def _get_callback_per_line_bytes_async(self):
        return self._callback_per_line_bytes_async

    def _set_callback_per_line_bytes_async(self, callback_per_line_bytes_async):
        if callback_per_line_bytes_async is None:
            self._callback_per_line_bytes_async = None
        elif asyncio.iscoroutinefunction(callback_per_line_bytes_async):
            self._callback_per_line_bytes_async = callback_per_line_bytes_async
        else:
            raise TypeError('A coroutine function or None is '
                            'required')

    callback_per_line_bytes_async = property(
        _get_callback_per_line_bytes_async, _set_callback_per_line_bytes_async)
    """Set callback coroutine function to get subprocess output bytes.

    """

    async def _process_stream_async(self, line, isstdout: bool = True):
        prefix = 'STDOUT'
        if not isstdout:
            prefix = 'STDERR'

        if isstdout:
            s = '{}pid={}: {}: {}'
            s = s.format(self.log_prefix, self.pid, prefix, line)
            self._logger.debug(s)
        else:
            s = '{}pid={}: {}: {}'
            s = s.format(self.log_prefix, self.pid, prefix, line)
            self._logger.info(s)
        # print('{}pid={}: {}: {}'.format(self.log_prefix, self.pid, prefix, line))

        if self._callback_per_line_bytes is not None:
            self._callback_per_line_bytes(line, isstdout)
        if self._callback_per_line is not None:
            self._callback_per_line(line.decode('utf-8'), isstdout)
        if self._callback_per_line_bytes_async is not None:
            await self._callback_per_line_bytes_async(line, isstdout)
        if self._callback_per_line_async is not None:
            await self._callback_per_line_async(line.decode('utf-8'), isstdout)

    async def _read_stream_async(self, stream, cb_async):
        futures = []
        while True:
            line = await stream.readline()
            if line:
                # call _adapt_stdout_async or _adapt_stderr_async
                f = asyncio.ensure_future(cb_async(line))
                futures.append(f)
            else:
                break
        if len(futures) > 0:
            await asyncio.wait(futures)

    async def _adapt_stdout_async(self, line):
        f = asyncio.ensure_future(self._process_stream_async(line, True))
        await f

    async def _adapt_stderr_async(self, line):
        f = asyncio.ensure_future(self._process_stream_async(line, False))
        await f

    async def exec(self):
        """Execute sub-process and returns code.

        This method is a coroutine.

        Returns
        -------
        int
            Return code of sub-process

        """
        create = asyncio.create_subprocess_exec(
            self.program, *self.args,
            stdout=asyncio.subprocess.PIPE,
            stderr=asyncio.subprocess.PIPE
        )
        # s = '{}launching process'
        # self._logger.info(s.format(self.log_prefix))
        proc = await create
        proc = cast_proc(proc)
        # self._set_pid(proc.pid)
        self.pid = proc.pid
        s = '{}pid={}: Process Start: program={}, args={}'
        self._logger.info(s.format(self.log_prefix, self.pid, self.program, str(self.args)))

        co_read_stdout = self._read_stream_async(
            proc.stdout, self._adapt_stdout_async)
        co_read_std_err = self._read_stream_async(
            proc.stderr, self._adapt_stderr_async)
        co_proc_itself = proc.wait()

        f_read_stdout = asyncio.ensure_future(co_read_stdout)
        f_read_stderr = asyncio.ensure_future(co_read_std_err)
        f_proc_itself = asyncio.ensure_future(co_proc_itself)

        futures = [f_read_stdout, f_read_stderr, f_proc_itself]
        return_when = concurrent.futures.ALL_COMPLETED

        self._process = proc
        await asyncio.wait(futures, return_when=return_when)
        self._process = None

        return_code = proc.returncode
        self._set_return_code(return_code)
        s = '{}pid={}: return code {}'
        s = '{}pid={}: Process End: return_code={}, program={}, args={}'
        s = s.format(self.log_prefix, self.pid, return_code, self.program, str(self.args))
        self._logger.info(s)
        return return_code

    def kill(self) -> None:
        """Kill running process.

        Call kill method(signal.SIGKILL) of the process(asyncio.subprocess.Process).
        """
        if self._process is not None:
            self._process.kill()

    def terminate(self) -> None:
        """Terminate running process.

        Call terminate method(signal.SIGTERM) of the process(asyncio.subprocess.Process).
        """
        if self._process is not None:
            self._process.terminate()

    def send_signal(self, signal_: int) -> None:
        """Send signal to the running process.

        Call send_signal method of the process(asyncio.subprocess.Process).
        """
        if self._process is not None:
            self._process.send_signal(signal_)


class SubProcessStdoutReceiver:
    """STDOUT receiver class for SubProcess.

    Gets whole output of SubProcess.

    #. Instantiate This
    #. Set this.callback() to SubProcess.callback_per_line
    #. Use whole STDOUT

        #. Use STDOUT as binary: this.get()
        #. Use STDOUT as string: this.get_string()

    """

    def __init__(self):
        """Init."""
        self._buffer = b''

    def callback(self, line, isstdout):
        """Store received data."""
        if isstdout:
            self._buffer += line

    def print(self):
        """Print all received data to console."""
        print(self._buffer)

    def print_string(self):
        """Print all received data as a utf-8 string to console."""
        print(self._buffer.decode('utf-8'))

    def get(self) -> str:
        """Get all received data."""
        return self._buffer

    def get_string(self) -> str:
        """Get all received data as a utf-8 string."""
        return self._buffer.decode('utf-8')



#!/usr/bin/env python
# -*- coding: utf-8 -*0
"""Utilities."""


import os
import shutil
import subprocess


class Du:
    """Find and keep du command file path."""

    def __init__(self):
        """Initialize self."""
        self._path: str = ''

    @property
    def path(self) -> str:
        """Get qemu-img command file path.

        Returns
        -------
        str
            du command file path.

        Raises
        ------
        DuNotFoundException
            If shutil.which('du') returns None
        """
        if len(self._path) > 0:
            return self._path
        path_ = shutil.which('du')
        if path_ is None:
            raise DuNotFoundException()
        self._path = path_
        return self._path

    @path.setter
    def path(self, path_: str):
        """Force the result to the value.

        Set a string to force the result of self.path.

        If you set empty string, self.path will re-find the du path.

        Parameters
        ----------
        path_ : str
        """
        self._path = path_

    def get_usage(self, path_: str) -> int:
        """Get used amount of the file or directory.

        Parameters
        ----------
        path_ : str

        Returns
        -------
        int : used amount(bytes) of the file or directory
            Returns -1 if any error occurred.
        """
        if os.path.exists(path_) is False:
            return -1
        val = subprocess.check_output([self.path, '-s', '--block-size', '1', path_])
        val_split = val.split()
        val_bytes = val_split[0]
        val_str = val_bytes.decode('utf-8')
        val_int = int(val_str)
        return val_int


class DuNotFoundException(Exception):
    """An exception when du command not found in the PATH."""

    pass


def is_root() -> bool:
    """Return True if current user is root."""
    return True if os.getuid() == 0 else False


def get_mount_point_from_path(path_: str) -> str:
    """Find device mount point which contains specified path.

    Parameters
    ----------
    path_
        target path(file or dir.)

    Returns
    -------
        the device mount point which contains target path
    """
    # normalize
    current_path = os.path.realpath(path_)
    current_path = os.path.normcase(current_path)

    # query parent path and device id
    current_device_id = os.stat(current_path).st_dev
    parent_path = os.path.dirname(current_path)
    parent_device_id = os.stat(parent_path).st_dev

    # parent_path == current_path: top level dir.
    # parent_dev._id != current_dev._id:
    #     current_path is the mount point of the device

    while parent_path != current_path and \
            parent_device_id == current_device_id:
        # search upper dir.
        current_path = parent_path
        parent_path = os.path.dirname(current_path)
        current_device_id = parent_device_id
        parent_device_id = os.stat(parent_path).st_dev

    return current_path


def get_device_name_from_path(path_: str) -> str:
    """Get device name contains specified path.

    This function relies on Unix (uses /proc/mounts).

    Parameters
    ----------
    path_

    Returns
    -------
    device name contains specified path
    If path is invalid or cannot search, returns empty string.
    """
    mount_path = get_mount_point_from_path(path_)
    try:
        with open('/proc/mounts', 'r') as f:
            for s in f:
                p = s.rstrip('\n').split()
                if p[1] == mount_path:
                    return p[0]
    except EnvironmentError:
        pass
    return ''


def estimate_absolute_file_path(path_: str) -> str:
    """Estimate absolute file path from specified path string.

    This function searches an existing file from path.
    Unlike usual solutions, this function works for 'zipapp' relative path.

    At the case below:

        /opt/yourapp/

        * your_zipapp.zip
        * datafile.txt

    You can easily get full path of the 'datafile.txt'

    >>> print(estimate_absolute_file_path('datafile.ini'))
    /opt/yourapp/datafile.txt

    Parameters
    ----------
    path_ : str
        Relative path.

    Returns
    -------
    str
        The full path of the file.
        Returns empty string if founds no file.

    Examples
    --------
    ex1. based on relative path from program

        * /tmp/program.py
        * /tmp/a.txt

    >>> print(estimate_absolute_file_path('a.txt'))
    '/tmp/a.txt'

    ex2. based on relative path from current directory

        * /tmp/program.py
        * /foo/a.txt
        * $ cd /foo
        * $ /tmp/program.py

    >>> print(estimate_absolute_file_path('a.txt'))
    '/foo/a.txt'

    ex3. absolute path itself

        * /tmp/a.txt

    >>> print(estimate_absolute_file_path('/tmp/a.txt'))
    '/tmp/a.txt'

    """
    if os.path.isabs(path_):
        if os.path.isfile(path_):
            return path_
        return ''

    # relative path from current directory
    relative_path = os.path.join(os.path.abspath(os.getcwd()), path_)
    if os.path.isfile(relative_path):
        return path_

    # relative path of zipapp
    dir_ = os.path.dirname(os.path.abspath(__file__))
    dir_ = os.path.join(dir_, '..')
    relative_path = os.path.join(dir_, path_)
    if os.path.isfile(relative_path):
        return relative_path

    # relative path of normal script / zipapp internal
    dir_ = os.path.dirname(os.path.abspath(__file__))
    relative_path = os.path.join(dir_, path_)
    if os.path.isfile(relative_path):
        return relative_path

    return ''



#!/usr/bin/env python
# -*- coding: utf-8 -*0
"""Dump to know what it is."""

import inspect
import sys
import typing

"""Callback function definition as dump_class

Parameters
----------
target : object
    The target object you want to dump.
indent : int
    Depth to indent
no_print : bool
    If False(Default), immediately output by print().
    If True, no output. Use return value.
public_only : bool
    Ignore methods starts with '_' when True.
Returns
-------
str
    Dump result text.
"""
DumpClassCallback = typing.Callable[[object, int, bool, bool], str]


def dump_class(target: object,
               indent: int = 0,
               no_print: bool = False,
               public_only: bool = True
               ) -> str:
    """Dump class information.

    Parameters
    ----------
    target : object
        The target object you want to dump.
    indent : int
        Depth to indent
    no_print : bool
        If False(Default), immediately output by print().
        If True, no output. Use return value.
    public_only : bool
        Ignore methods starts with '_' when True.

    Returns
    -------
    str
        Dump result text.

    """
    tp: type

    if inspect.isclass(target):
        tp = typing.cast(type, target)
    else:
        tp = target.__class__

    s = 'description of {}.{}:'
    s = '  ' * indent + s.format(tp.__module__, tp.__name__)
    dump_string = s + '\n'
    s = '  ' * (indent + 1) + '__hash__: {}'.format(_get_hash_string(tp))
    dump_string += s + '\n'
    s = '  ' * (indent + 1) + '__doc__: {}'.format(tp.__doc__)
    dump_string += s + '\n'
    s = '  ' * (indent + 1) + '__qualname__: {}'.format(tp.__qualname__)
    dump_string += s + '\n'
    s = '  ' * (indent + 1) + '__module__: {}'.format(tp.__module__)
    dump_string += s + '\n'
    s = '  ' * (indent + 1) + '__name__: {}'.format(tp.__name__)
    dump_string += s + '\n'

    s = 'methods of {}.{}:'
    s = '  ' * indent + s.format(tp.__module__, tp.__name__)
    dump_string += s + '\n'
    s = dump_methods(target, indent + 1, True, public_only)
    s = s.rstrip('\n')
    dump_string += s

    if no_print is False:
        print(dump_string)
    return dump_string + '\n'


def dump_class_simple(target: object,
                      indent: int = 0,
                      no_print: bool = False,
                      public_only: bool = True) -> str:
    """Dump class information (simple).

    A function like dump_class but no output of methods.

    Parameters
    ----------
    target : object
        The target object you want to dump.
    indent : int
        Depth to indent
    no_print : bool
        If False(Default), immediately output by print().
        If True, no output. Use return value.
    public_only : bool
        Ignore methods starts with '_' when True.

    Returns
    -------
    str
        Dump result text.

    """
    tp: type
    if inspect.isclass(target):
        tp = typing.cast(type, target)
    else:
        tp = target.__class__

    s = 'description of {}.{}:'
    s = '  ' * indent + s.format(tp.__module__, tp.__name__)
    dump_string = s + '\n'
    s = '  ' * (indent + 1) + '__hash__: {}'.format(_get_hash_string(tp))
    dump_string += s + '\n'
    s = '  ' * (indent + 1) + '__doc__: {}'.format(tp.__doc__)
    dump_string += s + '\n'
    s = '  ' * (indent + 1) + '__qualname__: {}'.format(tp.__qualname__)
    dump_string += s + '\n'
    s = '  ' * (indent + 1) + '__module__: {}'.format(tp.__module__)
    dump_string += s + '\n'
    s = '  ' * (indent + 1) + '__name__: {}'.format(tp.__name__)
    dump_string += s

    if no_print is False:
        print(dump_string)
    return dump_string + '\n'


def dump_class_dummy(target: object,
                     indent: int = 0,
                     no_print: bool = False,
                     public_only: bool = True) -> str:
    """Pseudo implementation of DumpClassCallback outputs and returns nothing.

    Parameters
    ----------
    target: object
    indent: int
    no_print: bool
    public_only : bool

    Returns
    -------
    str
        Always empty string

    """
    return ''


def dump_class_tree(target: object,
                    max_depth: int = 10,
                    no_print: bool = False,
                    indent: int = 0,
                    dump_class_callback: DumpClassCallback = dump_class_dummy,
                    public_only: bool = True) -> str:
    """Dump class information, and its ancestors.

    Dumps class tree only by default.

    Set dump_class() or any to dump_class_callback
    to dump with descriptions of the class.
    Or use dump_class_tree_verbose().

    Parameters
    ----------
    target : object
        The target you want to dump.
    max_depth : int, default 10
        Maximum number of ascending.
        Smaller number suppresses output simple.
    no_print: bool, default False
        Set True suspends all print() call.
        Get dump result from return value.
    indent: int, default 0
        Indent result text.
    dump_class_callback: DumpClassCallback, default dump_class_dummy
        Set output function for classes.
        Default function outputs nothing.
        Set dump_class_simple or dump_class
        to get detail for classes.
    public_only : bool
        Ignore methods starts with '_' when True.

    Returns
    -------
    str
        Dump result text.
    """
    return _dump_class_tree_internal(
        target,
        max_depth,
        indent,
        None,
        dump_class_callback,
        no_print,
        public_only)


def dump_class_tree_verbose(target: object,
                            max_depth: int = 10,
                            no_print: bool = False,
                            indent: int = 0,
                            public_only: bool = True) -> str:
    """Dump class information, and its ancestors.

    Unlike dump_class_tree(), this uses dump_class().
    (Default class dumper of dump_class_tree() is dump_class_dummy(),
    which dumps nothing.)

    Parameters
    ----------
    target : object
        The target you want to dump.
    max_depth : int, default 10
        Maximum number of ascending.
        Smaller number suppresses output simple.
    no_print: bool, default False
        Set True suspends all print() call.
        Get dump result from return value.
    indent: int, default 0
        Indent result text.
    public_only : bool
        Ignore methods starts with '_' when True.

    Returns
    -------
    str
        Dump result text.
    """
    return _dump_class_tree_internal(
        target,
        max_depth,
        indent,
        None,
        dump_class,
        no_print,
        public_only)


def _dump_class_tree_internal(target: object, depth: int = 10, indent: int = 0,
                              duplication_checker: dict = None,
                              dump_class_callback: DumpClassCallback = dump_class,
                              no_print: bool = False,
                              public_only: bool = True):
    dump_string = ''

    if depth <= 0:
        return dump_string

    if inspect.isclass(target) is False:
        target = target.__class__
    target = typing.cast(type, target)

    if duplication_checker is None:
        duplication_checker = dict()

    target_hash = _get_hash_string(target)
    if target_hash == '':
        target_hash = target.__module__ + '.' + target.__name__
    if target_hash in duplication_checker:
        # already dumped class
        duplication_checker[target_hash] += 1
        return dump_string

    duplication_checker[target_hash] = 0

    s = '  ' * indent + '{}.{}:'.format(target.__module__, target.__name__)
    dump_string += s + '\n'
    if no_print is False:
        print(s)

    dump_string += dump_class_callback(target, indent + 1, no_print, public_only)

    target_bases: list = list()
    if hasattr(target, '__bases__') and target.__bases__:
        target_bases = list(target.__bases__)
        try:
            target_bases.remove(object)
        except ValueError:
            pass

    if len(target_bases) > 0:
        s = '{} parent(s) found for [{}.{}]'
        s = s.format(len(target_bases), target.__module__, target.__name__)
        s = '  ' * (indent + 1) + s
        dump_string += s + '\n'
        if no_print is False:
            print(s)
        for c in target_bases:
            s = '  ' * (indent + 2) + '[{}.{}]'.format(c.__module__, c.__name__)
            dump_string += s + '\n'
            if no_print is False:
                print(s)

        for c in target_bases:
            c_hash = _get_hash_string(c)
            if c_hash in duplication_checker:
                # already dump class
                # s = '  '*(indent+2) + 'skipped - already dumped'
                # dump_string += s + '\n'
                # if no_print is False:
                #     print(s)
                duplication_checker[c_hash] += 1
                continue
            dump_string += _dump_class_tree_internal(
                c, depth - 1, indent + 1, duplication_checker,
                dump_class_callback,
                no_print,
                public_only)
    else:
        s = '  ' * (indent + 1) + 'have no parents'
        dump_string += s + '\n'
        if no_print is False:
            print(s)

    return dump_string


def _get_hash_string(target):
    """Calculate Object hash.

    Parameters
    ----------
    target
        An object. It must have __hash__ method.

    Returns
    -------
        The hash string.

    """
    try:
        a = target.__hash__(target)
        b = a % ((sys.maxsize + 1) * 2)
    except TypeError:
        return ''

    return str(b)


def dump_methods(target, indent: int = 0, no_print: bool = False, public_only: bool = True):
    """Dump methods of a class.

    Parameters
    ----------
    target : object
        The target you want to dump. Class or Instance is acceptable.
    indent : int
        Depth to indent
    no_print : bool
        If False(Default), immediately output by print().
        If True, no output. Use return value.
    public_only : bool
        Ignore methods starts with '_' when True.

    Returns
    -------
    str
        Dump result text.

    """
    dump_string = ''

    if False is inspect.isclass(target) \
            and isinstance(target, object):
        target = target.__class__
    if False is inspect.isclass(target):
        return dump_string

    for a in inspect.classify_class_attrs(target):
        if a.kind == 'method':
            attr = getattr(target, a.name)
            # target_class = target
            # if a.defining_class:
            #     target_class = a.defining_class
            attr = getattr(a.defining_class, a.name)
            signature = None
            try:
                signature = inspect.signature(attr)
            except ValueError:
                pass
            s = a.name
            if public_only:
                if s.startswith('_'):
                    continue
            if signature is not None:
                s += str(signature)

            if len(dump_string) > 0:
                dump_string += '\n'
            s = '  ' * indent + s
            dump_string += s

    if no_print is False:
        print(dump_string)

    return dump_string + '\n'


if __name__ == '__main__':
    print(dump_class_tree(True))



"""zobepy - zobe's unsorted library.

This is an unsorted library, so some of functions or classes
may move to separate packages in the future.

Shortnames
==========

You may use short names instead of poor long names.

.. csv-table::
    :header: Short Name, Long Name
    :widths: 5, 5

    zobepy.BinarySizeFormatter, zobepy.binary_size_formatter.BinarySizeFormatter
    zobepy.SubProcess, zobepy.subpr.SubProcess

"""

__version_info__ = ('1', '0', '4')
__version__ = '.'.join(__version_info__)

from .binary_size_formatter import BinarySizeFormatter
from .subpr import SubProcess, SubProcessStdoutReceiver
import zobepy.dump
import zobepy.util



#!/usr/bin/env python
# -*- coding: utf-8 -*0


"""A formatter from int to bit size or byte size expression."""


class BinarySizeFormatter:
    """Easily make number for binary size into eye-friendly string.

    Instantiate with number and
    call methods::

        x = zobepy.BinarySizeFormatter(123)
        print(x.get())
        # output string: 123 B
        print(x)
        # output string: 123 B

    Use default
    automatically chosen unit::

        x = zobepy.BinarySizeFormatter(12300)

        print(x.get())
        # output string: 12 KiB
        print(x.get_unit_string())
        # output string: KiB

    .. csv-table:: examples
        :header: "integer value", "converted text (default)"
        :widths: 5, 5

        123, "123 B"
        12300, "12 KiB"
        123000000, "117 MiB"
        1230000000000, "1.1 TiB"
        123000000000000000000000000, "102 YiB"
        12300000000000000000000000000000, "10,174,322 YiB"


    Note:
        The symbols and names are declared in
        IEC 60027-2 A.2 and ISO/IEC 80000.

        Reference: https://en.wikipedia.org/wiki/Binary_prefix

    Parameters
    ----------
    val : int
        An int value.

    """

    def __init__(self, val: int):
        """Initialize this instance."""
        self._val = val
        self._decide_preferred_unit()

    UNIT_B = 0
    """Unit 'bytes'. """

    UNIT_KIB = 1
    """Unit 'kibi bytes', 'KiB'"""

    UNIT_MIB = 2
    """Unit 'mebi bytes', 'MiB'"""

    UNIT_GIB = 3
    """Unit 'gibi bytes', 'GiB'"""

    UNIT_TIB = 4
    """Unit 'tebi bytes', 'TiB'"""

    UNIT_PIB = 5
    """Unit 'pebi bytes', 'PiB'"""

    UNIT_EIB = 6
    """Unit 'exbi bytes', 'EiB'"""

    UNIT_ZIB = 7
    """Unit 'zebi bytes', 'ZiB'"""

    UNIT_YIB = 8
    """Unit 'yobi bytes', 'YiB'"""

    UNIT_MAX = UNIT_YIB
    """Current largest unit symbol is YiB."""

    UNIT_MIN = UNIT_B
    """No symbol."""

    UNIT_SYMBOLS = {
        0: '',
        1: 'Ki',
        2: 'Mi',
        3: 'Gi',
        4: 'Ti',
        5: 'Pi',
        6: 'Ei',
        7: 'Zi',
        8: 'Yi',
    }
    """Binary prefix symbols declared in ISO/IEC 80000."""

    UNIT_NAMES = {
        0: '',
        1: 'kibi',
        2: 'mebi',
        3: 'gibi',
        4: 'tebi',
        5: 'pebi',
        6: 'exbi',
        7: 'zebi',
        8: 'yobi',
    }
    """Binary prefix names declared in ISO/IEC 80000."""

    UNIT_CUSTOMARY_SYMBOLS = {
        0: '',
        1: 'K',
        2: 'M',
        3: 'G',
        4: 'T',
        5: 'P',
        6: 'E',
        7: 'Z',
        8: 'Y',
    }
    """Customary binary prefixes."""

    UNIT_CUSTOMARY_NAMES = {
        0: '',
        1: 'kilo',
        2: 'mega',
        3: 'giga',
        4: 'tera',
        5: 'peta',
        6: 'exa',
        7: 'zeta',
        8: 'yotta',
    }
    """Customary binary prefix names declared in ISO/IEC 80000."""

    @staticmethod
    def round(val, digit: int = 0):
        """Round specified digit of the val.

        Parameters
        ----------
        val : int
            An int value.
        digit : int, default 0
            The digit to be rounded at.

        Returns
        -------
        float
            Rounded value.

        """
        p = 10 ** digit
        return (val * p * 2 + 1) // 2 / p

    @staticmethod
    def format(val: int) -> str:
        """Easily make number for bytes into eye-friendly string.

        The most simple way to use this class.

        Parameters
        ----------
        val : int
            An int value.

        Returns
        -------
        int
            Formatted string.

        """
        return str(BinarySizeFormatter(val))

    @staticmethod
    def format_base1000(val: int) -> str:
        """Easily make number for bytes into eye-friendly string.

        The most simple way to use this class.

        Unlike format(), this method treats val as decimal.
        ex. format() formats 1024 into 1 KiB
        ex. format_base1000() formats 1000 into 1 KB

        Parameters
        ----------
        val : int
            An int value.

        Returns
        -------
        int
            Formatted string.

        """
        return BinarySizeFormatter(val).get_base1000()

    def __str__(self):
        """Convert value into string with appropriate unit prefix."""
        return self.get()
        # return self.get_value_string() + ' ' + self.get_unit_string()

    def get(self, base1024: int = -1) -> str:
        """Convert value into string. You can choose unit prefix.

        .. csv-table::
            :header: "base1024", "description"

            -1, "default, unit prefix automatically selected"
            0, "expression in bytes. ex. 123: 123 B"
            1, "expression in KiB. ex. 1024: 1.0 KiB"
            2, "expression in MiB."
            "...", "..."
            8, "expression in YiB."

        Parameters
        ----------
        base1024 : int, default -1
            Force the unit.

        Returns
        -------
        str
            Formatted string of the value.

        See also
        --------
        get_value_string : get formatted string of numeric part only

        """
        return (self.get_value_string(base1024)
                + ' ' + self.get_unit_string(base1024))

    def get_base1000(self, base1000: int = -1) -> str:
        """Convert value into string. You can choose unit prefix.

        Unlike get(), this method calculates by decimal.
        ex. get() for int 1024: 1 KiB
        ex. get_base1000() for int 1000: 1 KB

        .. csv-table::
            :header: "base1000", "description"

            -1, "default, unit prefix automatically selected"
            0, "expression in bytes. ex. 123: 123 B"
            1, "expression in KB. ex. 1000: 1.0 KiB"
            2, "expression in MB."
            "...", "..."
            8, "expression in YB."

        Parameters
        ----------
        base1000 : int, default -1
            Force the unit.

        Returns
        -------
        str
            Formatted string of the value.

        See also
        --------
        get_value_string_base1000 : get formatted string of numeric part only

        """
        return (self.get_value_string_base1000(base1000)
                + ' ' + self.get_unit_string_base1000(base1000))

    def get_value_string(self, base1024: int = -1) -> str:
        """Convert numeric value into string.

        Returns number only.
        You can also use unit string getter methods.

        Parameters
        ----------
        base1024 : int, default -1
            Force the unit.
            Set 1 to force the unit 'KiB'.
            Set 2 to force the unit 'MiB'.
            Set 0 to force the unit none.
            Set -1 (default) to automatic.

        Returns
        -------
        str
            Formatted string of the numeric part of the value.

        See also
        --------
        get_unit_string : Like 'KiB', 'MiB'.
        get_unit_symbol : Like 'Ki', 'Mi'.
        get_unit_name : Like 'kibi', 'mebi'.
        get_unit_customary_symbol : Like 'K', 'M'.
        get_unit_customary_name : Like 'kilo', 'mega'.

        """
        if base1024 == -1:
            base1024 = self._unit

        if base1024 == 0:
            s = '{:,}'
            s = s.format(self._val)
            return s
        else:
            if self._val < (100 * (1024 ** (base1024 - 1))):
                return '0.0'
            elif self._val < (10 * (1024 ** base1024)):
                v = self._val / 1024 ** base1024
                v = BinarySizeFormatter.round(v, 1)
                return str(v)
            elif self._val < (100 * (1024 ** base1024)):
                v = self._val / 1024 ** base1024
                v = BinarySizeFormatter.round(v)
                v = int(v)
                return str(v)
            else:
                v = self._val / 1024 ** base1024
                v = BinarySizeFormatter.round(v)
                v = int(v)
                return '{:,}'.format(v)

    def get_value_string_base1000(self, base1000: int = -1) -> str:
        """Convert numeric value into string.

        Returns number only.
        You can also use unit string getter methods.

        Attention
        ---------
        Unlike get_value_string(), this method treats the value decimal(base 10).
        Unlike get_value_string() treats 1024 = 1 KiB,
        this method treats 1000 = 1 KB.
        It is convenient and correct for hard drive vendors.


        Parameters
        ----------
        base1000 : int, default -1
            Force the unit.
            Set 1 to force the unit 'KB'.
            Set 2 to force the unit 'MB'.
            Set 0 to force the unit none.
            Set -1 (default) to automatic.

        Returns
        -------
        str
            Formatted string of the numeric part of the value.

        See also
        --------
        get_unit_string : Like 'KiB', 'MiB'.
        get_unit_symbol : Like 'Ki', 'Mi'.
        get_unit_name : Like 'kibi', 'mebi'.
        get_unit_customary_symbol : Like 'K', 'M'.
        get_unit_customary_name : Like 'kilo', 'mega'.

        """
        if base1000 == -1:
            base1000 = self._unit

        if base1000 == 0:
            s = '{:,}'
            s = s.format(self._val)
            return s
        else:
            if self._val < (100 * (1000 ** (base1000 - 1))):
                return '0.0'
            elif self._val < (10 * (1000 ** base1000)):
                v = self._val / 1000 ** base1000
                v = BinarySizeFormatter.round(v, 1)
                return str(v)
            elif self._val < (100 * (1000 ** base1000)):
                v = self._val / 1000 ** base1000
                v = BinarySizeFormatter.round(v)
                v = int(v)
                return str(v)
            else:
                v = self._val / 1000 ** base1000
                v = BinarySizeFormatter.round(v)
                v = int(v)
                return '{:,}'.format(v)

    def get_unit(self) -> int:
        """Return preferred unit number, in BinarySizeFormatter.UNIT_* constants.

        * Same as base-1024 value.
        * ex. KiB = 1
        * ex. MiB = 2

        Returns
        -------
        int

        """
        return self._unit

    def get_unit_base1000(self) -> int:
        """Return preferred unit number, in BinarySizeFormatter.UNIT_* constants.

        * Same as base-1000 value.
        * ex. KB = 1
        * ex. MB = 2

        Returns
        -------
        int

        """
        return self._unit_base1000

    def get_unit_string(self, base1024: int = -1):
        """Return unit string, KiB, MiB, etc.

        if base1024 = 0, it returns 'B' (no prefix).

        Returns
        -------
        string

        """
        return self.get_unit_symbol(base1024) + 'B'

    def get_unit_string_base1000(self, base1000: int = -1):
        """Return unit string, KB, MB, etc.

        if base1000 = 0, it returns 'B' (no prefix).

        Returns
        -------
        string

        """
        return self.get_unit_customary_symbol(base1000) + 'B'

    def get_unit_symbol(self, base1024: int = -1):
        """Return binary prefix of ISO/IEC 80000. ex:'Ki' for 1024.

        * ex1. 0 = '' (no prefix)
        * ex2. 1 = 'Ki' (kibi)
        * ex3. 2 = 'Mi' (mebi)

        If do not specify base1024, it returns automatically preferred prefix
        for the initial value of the instance.

        Returns
        -------
        string

        """
        unit = base1024
        if base1024 == -1:
            unit = self.get_unit()

        if unit < BinarySizeFormatter.UNIT_MIN:
            unit = BinarySizeFormatter.UNIT_MIN
        elif unit > BinarySizeFormatter.UNIT_MAX:
            unit = BinarySizeFormatter.UNIT_MAX

        return BinarySizeFormatter.UNIT_SYMBOLS.get(unit, '')

    def get_unit_name(self, base1024: int = -1):
        """Return binary prefix name instead of the symbol. ex:'kibi' for 1024.

        * ex1. 0 = '' (no prefix)
        * ex2. 1 = 'kibi'
        * ex3. 2 = 'mebi'

        If do not specify base1024, it returns automatically preferred prefix
        for the initial value of the instance.

        Returns
        -------
        string

        """
        unit = base1024
        if base1024 == -1:
            unit = self.get_unit()

        if unit < BinarySizeFormatter.UNIT_MIN:
            unit = BinarySizeFormatter.UNIT_MIN
        elif unit > BinarySizeFormatter.UNIT_MAX:
            unit = BinarySizeFormatter.UNIT_MAX

        return BinarySizeFormatter.UNIT_NAMES.get(unit, '')

    def get_unit_customary_symbol(self, base1024: int = -1):
        """Return customary binary prefix. ex: 'K' for 1024.

        * ex1. 0 = '' (no prefix)
        * ex2. 1 = 'K' (prefix of 'kilo-bytes' instead of 'kibi')
        * ex3. 2 = 'M' (prefix of 'mega-bytes' instead of 'mebi')

        If do not specify base1024, it returns automatically preferred prefix
        for the initial value of the instance.

        Returns
        -------
        string

        """
        unit = base1024
        if base1024 == -1:
            unit = self.get_unit()

        if unit < BinarySizeFormatter.UNIT_MIN:
            unit = BinarySizeFormatter.UNIT_MIN
        elif unit > BinarySizeFormatter.UNIT_MAX:
            unit = BinarySizeFormatter.UNIT_MAX

        return BinarySizeFormatter.UNIT_CUSTOMARY_SYMBOLS.get(unit, '')

    def get_unit_customary_name(self, base1024: int = -1):
        """Return customary binary prefix name. ex:'kilo' for 1024.

        * ex1. 0 = '' (no prefix)
        * ex2. 1 = 'kilo'
        * ex3. 2 = 'mega'

        If do not specify base1024, it returns automatically preferred prefix
        for the initial value of the instance.

        Returns
        -------
        string

        """
        unit = base1024
        if base1024 == -1:
            unit = self.get_unit()

        if unit < BinarySizeFormatter.UNIT_MIN:
            unit = BinarySizeFormatter.UNIT_MIN
        elif unit > BinarySizeFormatter.UNIT_MAX:
            unit = BinarySizeFormatter.UNIT_MAX

        return BinarySizeFormatter.UNIT_CUSTOMARY_NAMES.get(unit, '')

    def _decide_preferred_unit(self):
        self._unit = BinarySizeFormatter.UNIT_MAX
        for i in range(0, 1 + BinarySizeFormatter.UNIT_MAX):
            if self._val < 512 * (1024**i):
                self._unit = i
                break
        for i in range(0, 1 + BinarySizeFormatter.UNIT_MAX):
            if self._val < 500 * (1000**i):
                self._unit_base1000 = i
                break
